/*
 * Copyright © 2009 Christian Persch
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef ATRIL_DOCUMENT_H
#define ATRIL_DOCUMENT_H

#define __EV_ATRIL_DOCUMENT_H_INSIDE__

#include <libdocument/ev-annotation.h>
#include <libdocument/ev-async-renderer.h>
#include <libdocument/ev-attachment.h>
#include <libdocument/ev-backends-manager.h>
#include <libdocument/ev-document-attachments.h>
#include <libdocument/ev-document-factory.h>
#include <libdocument/ev-document-find.h>
#include <libdocument/ev-document-fonts.h>
#include <libdocument/ev-document-forms.h>
#include <libdocument/ev-document.h>
#include <libdocument/ev-document-images.h>
#include <libdocument/ev-document-info.h>
#include <libdocument/ev-document-layers.h>
#include <libdocument/ev-document-print.h>
#include <libdocument/ev-document-links.h>
#include <libdocument/ev-document-misc.h>
#include <libdocument/ev-document-security.h>
#include <libdocument/ev-document-thumbnails.h>
#include <libdocument/ev-document-transition.h>
#include <libdocument/ev-document-type-builtins.h>
#include <libdocument/ev-file-exporter.h>
#include <libdocument/ev-file-helpers.h>
#include <libdocument/ev-form-field.h>
#include <libdocument/ev-image.h>
#include <libdocument/ev-init.h>
#include <libdocument/ev-layer.h>
#include <libdocument/ev-link-action.h>
#include <libdocument/ev-link-dest.h>
#include <libdocument/ev-link.h>
#include <libdocument/ev-mapping-list.h>
#include <libdocument/ev-page.h>
#include <libdocument/ev-render-context.h>
#include <libdocument/ev-selection.h>
#include <libdocument/ev-transition-effect.h>
#include <libdocument/ev-version.h>

#undef __EV_ATRIL_DOCUMENT_H_INSIDE__

#endif /* !ATRIL_DOCUMENT_H */
